import unittest.mock as mock

import pytest

from yatbaf.types import InlineKeyboardMarkup
from yatbaf_menu import URL
from yatbaf_menu import Menu


@pytest.mark.asyncio
async def test_render_message():
    message = mock.AsyncMock(["answer"], answer=mock.AsyncMock())
    menu = Menu(title="menu", buttons=[URL("title", "https://example.com")])
    await menu._render(
        message,
        _title="title",
        _markup=InlineKeyboardMarkup([]),
    )
    message.answer.assert_awaited_once_with(
        text="title",
        reply_markup=InlineKeyboardMarkup([]),
        parse_mode=None,
    )


@pytest.mark.asyncio
async def test_render_callback_query():
    query = mock.AsyncMock(
        ["answer", "message", "__usrctx__"],
        answer=mock.AsyncMock(),
        message=mock.AsyncMock(),
        __usrctx__={},
    )
    menu = Menu(title="menu", buttons=[URL("title", "https://example.com")])
    await menu._render(
        query,
        _title="title",
        _markup=InlineKeyboardMarkup([]),
    )
    query.answer.assert_awaited_once()
    query.message.edit.assert_awaited_once_with(
        text="title",
        reply_markup=InlineKeyboardMarkup([]),
        parse_mode=None,
    )


@pytest.mark.asyncio
async def test_render_callback_query_noanswer():
    query = mock.AsyncMock(
        ["answer", "message", "__usrctx__"],
        answer=mock.AsyncMock(),
        message=mock.AsyncMock(),
        __usrctx__={"_noanswer": True},
    )
    menu = Menu(title="menu", buttons=[URL("title", "https://example.com")])
    await menu._render(
        query,
        _title="title",
        _markup=InlineKeyboardMarkup([]),
    )
    query.answer.assert_not_awaited()
    query.message.edit.assert_awaited_once_with(
        text="title",
        reply_markup=InlineKeyboardMarkup([]),
        parse_mode=None,
    )



from __future__ import annotations

__all__ = (
    "AbstractButton",
    "Action",
    "URL",
    "Submenu",
    "Back",
)

import asyncio
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING

from yatbaf.handler import Handler
from yatbaf.types import InlineKeyboardButton

from .filter import CallbackPayload

if TYPE_CHECKING:
    from collections.abc import Awaitable
    from collections.abc import Callable

    from yatbaf import OnCallbackQuery
    from yatbaf.types import CallbackQuery
    from yatbaf.typing import HandlerCallable

    from .menu import Menu
    from .payload import Payload
    from .typing import Query


class AbstractButton(ABC):
    __slots__ = ()

    @abstractmethod
    def _init(
        self, menu: Menu, router: OnCallbackQuery, payload: Payload, /
    ) -> None:
        pass

    @abstractmethod
    async def _build(self, q: Query, /) -> InlineKeyboardButton | None:
        pass


class BaseButton(AbstractButton):
    __slots__ = (
        "_title",
        "_show",
    )

    def __init__(
        self,
        *,
        title: str | Callable[[Query], Awaitable[str]],
        show: Callable[[Query], Awaitable[bool]] | None = None,
    ) -> None:
        self._title = title
        self._show = show

    def __repr__(self) -> str:
        title = "`dynamic`" if callable(t := self._title) else t
        return f"<{self.__class__.__name__}[{title=!r}]>"

    async def _get_title(self, q: Query, /) -> str:
        if callable(self._title):
            return await self._title(q)
        return self._title

    async def _is_visible(self, q: Query, /) -> bool:
        if self._show is not None:
            return await self._show(q)
        return True


class CallbackButton(BaseButton):
    __slots__ = (
        "_payload",
        "_payload_extra",
    )

    def __init__(
        self,
        title: str | Callable[[Query], Awaitable[str]],
        show: Callable[[Query], Awaitable[bool]] | None = None,
        payload: Callable[[Query], str] | None = None,
    ) -> None:
        super().__init__(
            title=title,
            show=show,
        )
        self._payload: str | None = None
        self._payload_extra = payload

    def __eq__(self, other: object) -> bool:
        return other is self or (  # yapf: disable
            isinstance(other, self.__class__) and (
                other._title == self._title
                and other._show == self._show
                and other._payload_extra == self._payload_extra
            )
        )

    async def _build(self, q: Query) -> InlineKeyboardButton | None:
        if not await self._is_visible(q):
            return None

        extra = "" if (f := self._payload_extra) is None else f(q)
        return InlineKeyboardButton(
            text=await self._get_title(q),
            callback_data=f"{self._payload}{extra}",
        )


class Action(CallbackButton):
    """This button does the action"""
    __slots__ = ("_action",)

    def __init__(
        self,
        title: str | Callable[[Query], Awaitable[str]],
        action: HandlerCallable[CallbackQuery],
        *,
        show: Callable[[Query], Awaitable[bool]] | None = None,
        payload: Callable[[Query], str] | None = None,
    ) -> None:
        """
        :param title: String or Callable which returns button title.
        :param action: Callable to run on click. Must be unique for the menu.
        :param show: *Optional.* Callable which returns visibility status.
        :param payload: *Optional.* Sync function, returned string will be used
            as `callback_data`.
        """
        super().__init__(
            title=title,
            show=show,
            payload=payload,
        )
        self._action = action

    def __eq__(self, other: object) -> bool:
        return (
            super().__eq__(other)
            and other._action == self._action  # type: ignore[attr-defined]
        )

    def _init(
        self, m: Menu, router: OnCallbackQuery, payload: Payload, /
    ) -> None:
        if self._payload is not None:
            raise ValueError(
                f"{self!r} button must be unique to the entire menu."
            )

        self._payload = payload.get()
        handler = Handler(
            fn=self._action,
            update_type=router.update_type,
            filters=[CallbackPayload(self._payload)],
        )
        if handler in router._handlers:
            raise ValueError(f"{self!r} `action` must be unique to menu.")

        router.add_handler(handler)


class URL(BaseButton):
    """This button will open URL"""
    __slots__ = ("_url",)

    def __init__(
        self,
        title: str | Callable[[Query], Awaitable[str]],
        url: str | Callable[[Query], Awaitable[str]],
        *,
        show: Callable[[Query], Awaitable[bool]] | None = None,
    ) -> None:
        """
        :param url: String or Callable which returns url.
        :param title: String or Callable which returns button title.
        :param show: *Optional.* Callable which returns visibility status.
        """
        super().__init__(
            title=title,
            show=show,
        )
        self._url = url

    def __eq__(self, other: object) -> bool:
        return isinstance(other, URL) and (  # yapf: disable
            other is self or (
                other._title == self._title
                and other._url == self._url
                and other._show == self._show
            )
        )

    def _init(self, m: Menu, r: OnCallbackQuery, p: Payload, /) -> None:
        pass  # nothing to do

    async def _get_url(self, q: Query, /) -> str:
        if callable(self._url):
            return await self._url(q)
        return self._url

    async def _build(self, q: Query, /) -> InlineKeyboardButton | None:
        if not await self._is_visible(q):
            return None

        async with asyncio.TaskGroup() as tg:
            title = tg.create_task(self._get_title(q))
            url = tg.create_task(self._get_url(q))

        return InlineKeyboardButton(
            text=title.result(),
            url=url.result(),
        )


class Submenu(CallbackButton):
    """This button will open next menu"""
    __slots__ = ("_menu",)

    def __init__(
        self,
        title: str | Callable[[Query], Awaitable[str]],
        menu: str,
        *,
        show: Callable[[Query], Awaitable[bool]] | None = None,
        payload: Callable[[Query], str] | None = None,
    ) -> None:
        """
        :param menu: Submenu name (see :class:`~yatbaf_menu.menu.Menu`).
        :param title: String or Callable which returns button title.
        :param show: *Optional.* Callable which returns visibility status.
        :param payload: *Optional.* Sync function, returned string will be used
            as `callback_data`.
        """
        super().__init__(
            title=title,
            show=show,
            payload=payload,
        )
        self._menu = menu

    def __repr__(self) -> str:
        title = "`dynamic`" if callable(t := self._title) else t
        menu = self._menu
        return f"<Submenu[{title=}, {menu=}]>"

    def __eq__(self, other: object) -> bool:
        return (
            super().__eq__(other)
            and other._menu == self._menu  # type: ignore[attr-defined]
        )

    def _init(self, menu: Menu, r: OnCallbackQuery, p: Payload, /) -> None:
        if self._payload is not None:
            raise ValueError(f"{self!r} button must be unique.")
        self._payload = f"{menu.get_submenu(self._menu)._prefix}##"


class Back(CallbackButton):
    """This button will open previous menu"""
    __slots__ = ()

    def __init__(
        self,
        title: str | Callable[[Query], Awaitable[str]],
        *,
        payload: Callable[[Query], str] | None = None,
    ) -> None:
        """
        :param title: String or Callable which returns button title.
        :param payload: *Optional.* Sync function, returned string will be used
            as `callback_data`.
        """
        super().__init__(
            title=title,
            payload=payload,
        )

    def _init(self, menu: Menu, r: OnCallbackQuery, p: Payload, /) -> None:
        if self._payload is not None:
            raise ValueError(f"{self!r} button must be unique.")

        if menu._parent is None:
            raise ValueError(f"It is not possible to use {self!r} in {menu!r}.")

        self._payload = f"{menu._parent._prefix}##"



from yatbaf_menu.button import Action
from yatbaf_menu.button import Back
from yatbaf_menu.button import Submenu
from yatbaf_menu.menu import Menu
from yatbaf_menu.row import Row
from yatbaf_menu.utils import build_router


async def action1(_):
    pass


async def action2(_):
    pass


def test_build():
    menu = Menu(
        title="title",
        name="main",
        buttons=[
            Submenu(title="submenu", menu="sub"),
            Action(title="button2", action=action2),
        ],
        submenu=[
            Menu(
                title="title",
                name="sub",
                buttons=[Action(title="button1", action=action1)],
                back_btn_title="back",
            )
        ],
    )

    router = build_router(menu)

    assert len(router._handlers) == 2  # base, fallback
    assert len(router._handlers[0]._handlers) == 2  # menu, submenu
    assert len(router._handlers[0]._handlers[0]._handlers) == 2  # menu, action
    assert len(router._handlers[0]._handlers[1]._handlers) == 2  # menu, action
    assert len(router._handlers[1]._handlers) == 1  # fallback action

    assert menu._parent is None
    assert menu._prefix is not None
    assert menu._rows == [
        Row([Submenu(title="submenu", menu="sub")]),
        Row([Action(title="button2", action=action2)]),
    ]

    submenu = menu._submenu["sub"]
    assert submenu._parent is menu
    assert submenu._prefix is not None
    assert submenu._rows == [
        Row([Action(title="button1", action=action1)]),
        Row([Back(title="back")]),
    ]



from __future__ import annotations

__all__ = ("build_router",)

from typing import TYPE_CHECKING
from typing import cast

from yatbaf import OnCallbackQuery
from yatbaf.middleware import Middleware

from .filter import CallbackPayload
from .filter import HasMessage
from .filter import MenuPrefix
from .filter import MenuVersion
from .filter import RootMenuPrefix
from .middleware import CutPayloadMiddileware
from .middleware import InjectMenuMiddleware
from .payload import Payload

if TYPE_CHECKING:
    from yatbaf.handler import BaseHandler
    from yatbaf.types import CallbackQuery

    from .menu import Menu
    from .nav import MenuNav

_root_payload_bucket = Payload()


def _build_menu_router(menu: Menu) -> OnCallbackQuery:
    """
    :meta private:

    :param menu: Menu object.
    """
    menu._prefix = cast("str", menu._prefix)
    payload = menu._prefix[4:6]

    menu._handler._filters = [CallbackPayload(f"{menu._prefix}##")]
    group = menu._group or OnCallbackQuery(name=menu.name)
    group.add_filter(MenuPrefix(payload))
    group.add_middleware(
        Middleware(InjectMenuMiddleware, menu=menu),
        scope="local",
    )
    group.add_handler(cast("BaseHandler[CallbackQuery]", menu._handler))
    return group


def _build_base_router(menu: Menu, version: str) -> OnCallbackQuery:
    """
    :meta private:

    :param menu: Main menu object.
    """
    return OnCallbackQuery(
        name=f"{menu.name}-root",
        filters=[MenuVersion(version)],
        handler_middleware=[CutPayloadMiddileware],
    )


def _build_fallback_router(menu: Menu, update_message: str) -> OnCallbackQuery:
    """
    :meta private:

    :param menu: Main menu object.
    :param update_message: Will be shown to the user.
    """
    router = OnCallbackQuery(
        name=f"{menu.name}-fallback",
        handler_middleware=[
            (  # type: ignore[list-item]
                Middleware(
                    InjectMenuMiddleware,  # type: ignore[arg-type]
                    menu=menu,
                ),
                "local",
            ),
        ],
    )

    @router
    async def _fallback(q: CallbackQuery) -> None:
        nav: MenuNav = q.ctx["nav"]
        q.data = ""
        await q.answer(text=update_message)
        nav.menu.noanswer(q)
        await nav.root()

    return router


def _build_root_router(
    menu: Menu,
    base_router: OnCallbackQuery,
    fallback_router: OnCallbackQuery,
) -> OnCallbackQuery:
    """:meta private:"""
    menu._prefix = cast("str", menu._prefix)
    prefix = menu._prefix[:2]  # root prefix 2 chars

    return OnCallbackQuery(
        name=f"{menu.name}-base",
        filters=[RootMenuPrefix(prefix), HasMessage()],
        handlers=[
            base_router,
            fallback_router,
        ],
    )


def _init_buttons(menu: Menu, router: OnCallbackQuery) -> None:
    menu._prefix = cast("str", menu._prefix)
    payload = Payload(menu._prefix)
    for row in menu._rows:
        row._init(menu, router, payload)


def _parse_version(version: str | int) -> str:
    version = str(version)
    if len(version) > 2:
        raise ValueError(f"{version=} cannot be longer than 2 characters.")
    return version.zfill(2)


def build_router(
    menu: Menu,
    *,
    version: str | int = 0,
    update_message: str = "This menu was updated.",
) -> OnCallbackQuery:
    """Build menu router.

    :param menu: :class:`~yatbaf_menu.menu.Menu` object.
    :param version: *Optional.* Menu version. Up to 2 chars.
    :param update_message: *Optional.* Will be shown to the user when
        ``version`` is increased.
    :returns: Configured :class:`~yatbaf.router.OnCallbackQuery` object.
    """
    version = _parse_version(version)
    menu_payload = Payload(_root_payload_bucket.get() + version)
    base_router = _build_base_router(menu, version)

    def _init_menu(menu: Menu) -> None:
        menu._prefix = menu_payload.get()
        router = _build_menu_router(menu)
        base_router.add_handler(router)
        for submenu in menu._submenu.values():
            submenu._parent = menu
            _init_menu(submenu)
        _init_buttons(menu, router)

    _init_menu(menu)

    return _build_root_router(
        menu=menu,
        base_router=base_router,
        fallback_router=_build_fallback_router(
            menu=menu,
            update_message=update_message,
        ),
    )



from __future__ import annotations

__all__ = ("CallbackPayload",)

from typing import TYPE_CHECKING
from typing import final

from yatbaf.filters import BaseFilter

if TYPE_CHECKING:
    from yatbaf.types import CallbackQuery
    from yatbaf.typing import FilterPriority


@final
class CallbackPayload(BaseFilter):
    __slots__ = ("payload",)

    def __init__(self, payload: str) -> None:
        self.payload = payload

    @property
    def priority(self) -> FilterPriority:
        return {"content": (1, 100)}

    async def check(self, q: CallbackQuery) -> bool:
        return q.data[:8] == self.payload  # type: ignore[index]


@final
class MenuVersion(BaseFilter):
    __slots__ = ("version",)

    def __init__(self, version: str) -> None:
        self.version = version

    @property
    def priority(self) -> FilterPriority:
        return {"content": (1, 100)}

    async def check(self, q: CallbackQuery) -> bool:
        # aa[00]bbcc
        return q.data[2:4] == self.version  # type: ignore[index]


@final
class MenuPrefix(BaseFilter):
    __slots__ = ("prefix",)

    def __init__(self, prefix: str) -> None:
        self.prefix = prefix

    @property
    def priority(self) -> FilterPriority:
        return {"content": (1, 100)}

    async def check(self, q: CallbackQuery) -> bool:
        # aa00[bb]cc
        return q.data[4:6] == self.prefix  # type: ignore[index]


@final
class RootMenuPrefix(BaseFilter):
    __slots__ = ("prefix",)

    def __init__(self, prefix: str) -> None:
        self.prefix = prefix

    @property
    def priority(self) -> FilterPriority:
        return {"content": (1, 100)}

    async def check(self, q: CallbackQuery) -> bool:
        # [aa]00bbcc
        return (d := q.data) is not None and d[:2] == self.prefix


RootMenuPrefix.incompat(MenuPrefix)


@final
class HasMessage(BaseFilter):
    __slots__ = ()

    @property
    def priority(self) -> FilterPriority:
        return {"content": (1, 100)}

    async def check(self, q: CallbackQuery) -> bool:
        return q.message is not None and q.message.date != 0



from __future__ import annotations

__all__ = ("Menu",)

import asyncio
from itertools import count
from typing import TYPE_CHECKING
from typing import cast
from typing import final

from yatbaf.di import Provide
from yatbaf.handler import Handler
from yatbaf.types import InlineKeyboardMarkup

from .button import Back
from .row import AbstractRowBuilder
from .row import Row

if TYPE_CHECKING:
    from collections.abc import Awaitable
    from collections.abc import Callable
    from collections.abc import Sequence

    from yatbaf.enums import ParseMode
    from yatbaf.group import OnCallbackQuery
    from yatbaf.types import CallbackQuery

    from .button import AbstractButton
    from .typing import Query

_menu_count = count(1).__next__


def _parse_markup(
    markup: list[AbstractButton | AbstractRowBuilder | list[AbstractButton]]
) -> list[AbstractRowBuilder]:
    result: list[AbstractRowBuilder] = []
    for row in markup:
        if isinstance(row, list):
            for obj in row:
                if issubclass(type(obj), AbstractRowBuilder):
                    if len(row) != 1:
                        raise ValueError(
                            f"It is not possible to use {obj!r} with "
                            "other objects in the same row"
                        )
                    result.append(cast("AbstractRowBuilder", obj))
                    break
            else:
                result.append(Row(row))

        # row builder
        elif issubclass(type(row), AbstractRowBuilder):
            result.append(cast("AbstractRowBuilder", row))

        # single button
        else:
            result.append(Row([cast("AbstractButton", row)]))

    return result


def _parse_menu_name(name: str | None) -> str:
    if name is not None:
        return name.lower()
    return f"menu{_menu_count()}"


@final
class Menu:
    """Menu object."""

    __slots__ = (
        "_title",
        "_name",
        "_buttons",
        "_submenu",
        "_rows",
        "_parent",
        "_prefix",
        "_group",
        "_handler",
        "_parse_mode",
    )

    def __init__(  # yapf: disable
        self,
        *,
        title: str | Callable[[Query], Awaitable[str]],
        name: str | None = None,
        buttons: list[AbstractButton | AbstractRowBuilder | list[AbstractButton]] | None = None,  # noqa: E501
        submenu: Sequence[Menu] | None = None,
        group: OnCallbackQuery | None = None,
        back_btn_title: str | Callable | None = None,
        parse_mode: ParseMode | None = None,
    ) -> None:
        """
        :param title: String or Callable which return menu title.
        :param name: *Optional.* Menu name.
        :param submenu: *Optional.* Sequence of :class:`Menu`.
        :param buttons: *Optional.* A list of buttons.
        :param group: *Optional.* Pre-configured handler group.
        :param back_btn_title: *Optional.* Pass a title if you want to add
            a 'back button' to this menu. For submenu only.
        :param parse_mode: *Optional.* Parse mode for menu title.
        """  # noqa: E501
        self._name = _parse_menu_name(name)
        self._title = title
        self._prefix: str | None = None

        buttons = buttons if buttons is not None else []
        if back_btn_title is not None:
            buttons.append(Back(title=back_btn_title))
        if not buttons:
            raise ValueError(f"{self!r} must have at least one button.")

        self._rows = _parse_markup(buttons)

        self._submenu = {} if submenu is None else {m.name: m for m in submenu}
        self._parent: Menu | None = None

        self._group = group
        self._handler = Handler(  # yapf: disable
            fn=self._render,
            update_type="callback_query",
            dependencies={
                "_title": Provide(title if callable(title) else self._get_title),  # noqa: E501
                "_markup": Provide(self._get_markup),
            }
        )
        self._parse_mode = parse_mode

    def __repr__(self) -> str:
        return f"<Menu[{self._name}]>"

    @property
    def name(self) -> str:
        """Menu name."""
        return self._name

    async def provide(self) -> Menu:
        return self

    def get_submenu(self, name: str) -> Menu:
        """Returns submenu by name.

        :param name: Submenu name.
        :raises ValueError: If submenu with ``name`` is not found.
        """
        try:
            return self._submenu[name]
        except KeyError:
            raise ValueError(f"Menu {name} not found in {self!r}") from None

    def get_root(self) -> Menu:
        """Returns main menu."""
        menu = self
        while menu._parent is not None:
            menu = menu._parent
        return menu

    def get_menu(self, path: str) -> Menu:
        """Returns menu by path.

        :param path: Path to menu.
        :raises ValueError: If ``path`` is not found.
        """
        menu = self.get_root()
        for name in path.split("."):
            menu = menu.get_submenu(name)
        return menu

    async def _get_title(self) -> str:
        """Menu title provider. Static title."""
        return self._title  # type: ignore[return-value]

    async def _get_markup(self, update: Query) -> InlineKeyboardMarkup:
        """Menu buttons provider."""
        if len(self._rows) == 1:
            return InlineKeyboardMarkup([
                row for row in await self._rows[0]._build(update)
            ])

        async with asyncio.TaskGroup() as tg:
            tasks = [tg.create_task(r._build(update)) for r in self._rows]
        return InlineKeyboardMarkup([
            row for task in tasks for row in task.result()
        ])

    async def _render(
        self,
        q: Query,
        _title: str,
        _markup: InlineKeyboardMarkup,
    ) -> None:
        """Menu handler callback.

        :param q: Update instance.
        :param _title: Menu title.
        :param _markup: Menu buttons.
        """
        # callback query
        if hasattr(q, "message"):
            if not q.__usrctx__.get("_noanswer", False):
                await q.answer()  # type: ignore[call-arg]
            # q.message is always a Message (filter.HasMessage)
            await q.message.edit(  # type: ignore[union-attr]
                text=_title,
                reply_markup=_markup,
                parse_mode=self._parse_mode,
            )

        # message
        else:
            await q.answer(
                text=_title,
                reply_markup=_markup,
                parse_mode=self._parse_mode,
            )

    async def render(self, q: Query, /) -> None:
        """Render menu."""
        await self._handler._exec(q)

    @staticmethod
    def noanswer(q: CallbackQuery, /) -> None:
        """Use this method before calling :meth:`render` to avoid calling
        `q.answer()` twice, if you use `q.answer()` manually.

        .. note::

            All methods of :class:`~yatbaf_menu.nav.MenuNav` use :meth:`render`.

        :param q: Callback event.
        """
        q.__usrctx__["_noanswer"] = True



from __future__ import annotations

__all__ = ("MenuNav",)

from typing import TYPE_CHECKING
from typing import cast

from yatbaf.exceptions import MethodInvokeError

if TYPE_CHECKING:
    from yatbaf.types import CallbackQuery
    from yatbaf.types import Message

    from .menu import Menu


class MenuNav:
    """Menu navigation.

    Use `update.ctx["nav"]` to access this object inside the button callback::

        async def button_action(q: CallbackQuery) -> None:
            nav = q.ctx["nav"]
            ...
    """

    __slots__ = (
        "_query",
        "_menu",
    )

    def __init__(self, menu: Menu, query: CallbackQuery, /) -> None:
        """
        :param menu: Current menu.
        :param query: Current query.
        """
        self._query = query
        self._menu = menu

    @property
    def menu(self) -> Menu:
        """Current :class:`~yatbaf_menu.menu.Menu` instance."""
        return self._menu

    async def navigate(self, path: str, /) -> None:
        """Use this method to open menu by path.

        :param path: Path relative to main menu. Separated by dots.
        """
        menu = self._menu.get_menu(path)
        await menu.render(self._query)

    async def back(self) -> None:
        """Use this method to go to previous menu."""
        menu = self._menu._parent
        if menu is None:
            return
        await menu.render(self._query)

    async def root(self) -> None:
        """Use this method to go to main menu."""
        menu = self._menu.get_root()
        await menu.render(self._query)

    async def submenu(self, name: str) -> None:
        """Use this method to go to submenu.

        :param name: Submenu name.
        """
        menu = self._menu.get_submenu(name)
        await menu.render(self._query)

    async def refresh(self) -> None:
        """Use this method to refresh current menu."""
        await self._menu.render(self._query)

    async def close(self) -> None:
        """Use this method to close the menu."""
        message = cast("Message", self._query.message)
        try:
            await message.delete()
        except MethodInvokeError:
            await message.edit_reply_markup()
